// Copyright 2004 Adam Megacz, see the COPYING file for licensing [GPL]
package org.ibex.plat;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.net.*;
import java.util.*;
import org.ibex.util.*;
import org.ibex.*;
import java.lang.reflect.*;


/** Platform class for most reasonable Java1.2+ Java2s */
public class Java2 extends AWT {

    private boolean isJava14 = false;

    public Java2() {
        // disable the focus manager so we can intercept the tab key
        String versionString = System.getProperty("java.version", "");
        int secondDecimal = versionString.substring(versionString.indexOf('.') + 1).indexOf('.');
        if (secondDecimal != -1) versionString = versionString.substring(0, versionString.indexOf('.') + 1 + secondDecimal);
        double version = Double.parseDouble(versionString);
        if (version >= 1.4) {
            isJava14 = true;
            try {
                Toolkit t = java.awt.Toolkit.getDefaultToolkit();
                Method m = java.awt.Toolkit.class.getMethod("setDynamicLayout", new Class[] { Boolean.TYPE });
                m.invoke(t, new Object[] { Boolean.TRUE });
            } catch (Exception e) {
                Log.info(this, "Exception while trying to enable AWT Dynamic Layout");
                Log.info(this, e);
            }
        }
        javax.swing.FocusManager.setCurrentManager(new javax.swing.FocusManager() {
                public void processKeyEvent(Component focusedComponent, KeyEvent anEvent) { }
                public void focusPreviousComponent(Component aComponent) { }
                public void focusNextComponent(Component aComponent) { }
            });
    }

    /** this is done with reflection in case a new version of the plugin comes out that doesn't let us pull the sun.plugin.* trick */
    protected synchronized org.ibex.HTTP.Proxy _detectProxy() {
        return (org.ibex.HTTP.Proxy)java.security.AccessController.doPrivileged(new java.security.PrivilegedAction() {
                public Object run() {
                    try {
                        org.ibex.HTTP.Proxy pi = new org.ibex.HTTP.Proxy();
                        
                        Class PluginProxyHandler = Class.forName("sun.plugin.protocol.PluginProxyHandler");
                        Method getDefaultProxyHandler = PluginProxyHandler.getMethod("getDefaultProxyHandler", new Class[] { });
                        Object proxyHandler = getDefaultProxyHandler.invoke(null, new Object[] { });
                        
                        Class ProxyHandler = Class.forName("sun.plugin.protocol.ProxyHandler");
                        Method getProxyInfo = ProxyHandler.getMethod("getProxyInfo", new Class[] { URL.class });
                        Object proxyInfo = getProxyInfo.invoke(proxyHandler, new Object[] { new URL("http://www.ibex.org") });
                        
                        Class ProxyInfo = Class.forName("sun.plugin.protocol.ProxyInfo");
                        
                        if (((Boolean)ProxyInfo.getMethod("isSocksUsed", new Class[] { }).invoke(proxyInfo, new Object[] { })).booleanValue()) {
                            pi.socksProxyHost =
                                (String)ProxyInfo.getMethod("getSocksProxy", new Class[] { }).invoke(proxyInfo, new Object[] { });
                            pi.socksProxyPort =
                                ((Integer)ProxyInfo.getMethod("getSocksPort", new Class[] { }).invoke(proxyInfo, new Object[] { })).intValue();
                        }
                        
                        if (((Boolean)ProxyInfo.getMethod("isProxyUsed", new Class[] { }).invoke(proxyInfo, new Object[] { })).booleanValue()) {
                            pi.httpProxyHost =
                                (String)ProxyInfo.getMethod("getProxy", new Class[] { }).invoke(proxyInfo, new Object[] { });
                            pi.httpProxyPort =
                                ((Integer)ProxyInfo.getMethod("getPort", new Class[] { }).invoke(proxyInfo, new Object[] { })).intValue();
                        }
                        
                        if (pi.httpProxyHost != null || pi.socksProxyHost != null) return pi;
                        else return null;

                    } catch (Throwable e) {
                        if (Log.on) Log.info(this, "No proxy information found in Java Plugin classes");
                        return null;
                    }
                }});
    }

    protected PixelBuffer _createPixelBuffer(int w, int h, Surface owner) { return new Java2PixelBuffer(w, h); }
    protected Surface _createSurface(final Box root, final boolean framed) {
        return (Surface)java.security.AccessController.doPrivileged(new java.security.PrivilegedAction() {
                public Object run() {
                    if (isJava14) {
                        try {
                             // weaken the binding here to avoid link errors on 1.3.x
                             Class java14SurfaceClass = Class.forName(Java2.class.getName() + "$Java14Surface");
                             Constructor ctor = java14SurfaceClass.getConstructor(new Class[] { Box.class, Boolean.TYPE });
                               return (Surface)ctor.newInstance(new Object[] { root, Boolean.valueOf(framed) });
                        } catch (Exception e) {
                            Log.info(this, e);
                            throw new LinkageError("error: " + e);
                        }
                    } else {
                        return new Java2Surface(root, framed);
                    }
                }
            });
    }

    // Inner Classes //////////////////////////////////////////////////////////////////

    private static Cursor invisibleCursor =
        Toolkit.getDefaultToolkit().createCustomCursor(new BufferedImage(2, 2, BufferedImage.TYPE_INT_ARGB),
                                                       new Point(1, 1), "invisible");

    protected static class Java2Surface extends AWTSurface {

        public Java2Surface(Box root, boolean framed) { super(root, framed); }

        protected void _setMinimized(boolean b) {
            if (frame == null) Log.info(this, "JDK 1.2 can only minimize frames, not windows");
            else if (b) frame.setState(java.awt.Frame.ICONIFIED);
            else frame.setState(java.awt.Frame.NORMAL);
        }

        public void syncCursor() {
            if (cursor.equals("invisible")) window.setCursor(invisibleCursor);
            else super.syncCursor();
        }
    }

    protected static class Java14Surface extends Java2Surface implements WindowStateListener, MouseWheelListener {
        public Java14Surface(Box root, boolean framed) {
            super(root, true);
            // JDK1.4 doesn't like java.lang.Window's...
            if (!framed) ((Frame)window).setUndecorated(true);
            window.addWindowStateListener(this);
            window.addMouseWheelListener(this);
            window.setVisible(true);
        }

        protected void makeVisible() { }
        
        protected void _setMaximized(boolean m) {
            if (frame == null) {
                if (Log.on) Log.info(this, "JDK 1.4 can only maximize frames, not windows");
                return;
            }
            frame.setExtendedState(m ? Frame.MAXIMIZED_BOTH : (minimized ? Frame.ICONIFIED : Frame.NORMAL));
        }
        protected void _setMinimized(boolean m) {
            if (frame == null) {
                if (Log.on) Log.info(this, "JDK 1.4 can only minimize frames, not windows");
                return;
            }
            frame.setExtendedState(m ? Frame.ICONIFIED : (maximized ? Frame.MAXIMIZED_BOTH : Frame.NORMAL));
        }
        public void windowStateChanged(WindowEvent e) {
            if (e.getOldState() != e.getNewState()) {
                if ((e.getNewState() & Frame.MAXIMIZED_BOTH) != 0) Maximized(true);
                else if (((e.getOldState() & Frame.MAXIMIZED_BOTH) != 0) && (e.getNewState() & Frame.MAXIMIZED_BOTH) == 0)
                    Maximized(false);
            }
        }

        public void mouseWheelMoved(MouseWheelEvent m) {
            if (m.getScrollType() == MouseWheelEvent.WHEEL_UNIT_SCROLL)
                VScroll(m.getUnitsToScroll());
        }
    }

    protected static class Java2PixelBuffer extends AWTPixelBuffer {
        private static ColorModel cm = Toolkit.getDefaultToolkit().getColorModel();
        private static Hashtable emptyHashtable = new Hashtable();
        private static short[] sbank = null;
        private static int[] ibank = null;
        private static byte[] bbank = null;
        private static int bank_start = 0;
        private WritableRaster raster = null;
        private SampleModel sm = null;
        private DataBuffer buf = null;

        // this doens't seem to work on Windows
        public void drawGlyph(Picture source, int dx, int dy, int cx1, int cy1, int cx2, int cy2, int rgb) {
            AWTPicture src = (AWTPicture)source;
            src.init();
            Graphics2D g2 = (Graphics2D)i.getGraphics();
            g2.setComposite(AlphaComposite.DstOut);
            g2.setClip(cx1, cy1, cx2 - cx1, cy2 - cy1);
            g2.drawImage(src.i, dx, dy, null);
            g2.setComposite(AlphaComposite.DstOver);
            g2.setColor(new Color((rgb & 0x00FF0000) >> 16, (rgb & 0x0000FF00) >> 8, (rgb & 0x000000FF)));
            g2.fillRect(dx, dy, cx2 - dx, cy2 - dy);
            g2.drawImage(i, 0, 0, null);
            g2.setClip(0, 0, i.getWidth(null), i.getHeight(null));
        }

        public Java2PixelBuffer(int w, int h) {
            sm = cm.createCompatibleSampleModel(w, h);
            int numSamples = w * h * sm.getNumDataElements();
            if (sm.getDataType() == DataBuffer.TYPE_USHORT) {
                if (sbank == null || numSamples > 512 * 512 / 3) {
                    buf = new DataBufferUShort(numSamples);
                } else {
                    if (numSamples > sbank.length - bank_start) {
                        bank_start = 0;
                        sbank = new short[512 * 512];
                    }
                    buf = new DataBufferUShort(sbank, numSamples, bank_start);
                    bank_start += numSamples;
                }
            } else if (sm.getDataType() == DataBuffer.TYPE_BYTE) {
                if (bbank == null || numSamples > 512 * 512 / 3) {
                    buf = new DataBufferByte(numSamples);
                } else {
                    if (numSamples > bbank.length - bank_start) {
                        bank_start = 0;
                        bbank = new byte[512 * 512];
                    }
                    buf = new DataBufferByte(bbank, numSamples, bank_start);
                    bank_start += numSamples;
                }
            } else if (sm.getDataType() == DataBuffer.TYPE_INT) {
                if (ibank == null || numSamples > 512 * 512 / 3) {
                    buf = new DataBufferInt(numSamples);
                } else {
                    if (numSamples > ibank.length - bank_start) {
                        bank_start = 0;
                        ibank = new int[512 * 512];
                    }
                    buf = new DataBufferInt(ibank, numSamples, bank_start);
                    bank_start += numSamples;
                }
            }
            raster = Raster.createWritableRaster(sm, buf, null);
            i = new BufferedImage(cm, raster, false,  emptyHashtable);
            g = i.getGraphics();
        }
    }

    protected String getDescriptiveName() { return isJava14 ? "Java 1.4+ JVM" : "Java 1.2+ JVM"; }

}
